from __future__ import annotations

import random
import shutil
from contextlib import contextmanager
from dataclasses import dataclass
from hashlib import md5
from io import BytesIO, StringIO
from operator import attrgetter
from pathlib import Path
from typing import (
    Generator,
    Generic,
    Literal,
    Optional,
    Protocol,
    Sequence,
    TypeVar,
    Union,
)
from urllib.parse import urlparse
from zipfile import Path as ZipPath
from zipfile import ZipFile

import orjson
import pandas as pd
from ruamel.yaml import YAML
from slugify import slugify
from typing_extensions import Self

import serde.csv
import serde.yaml
from serde import json
from sm.inputs.prelude import ColumnBasedTable, Context, Link
from sm.misc.funcs import batch
from sm.misc.matrix import Matrix
from sm.namespaces.namespace import Namespace
from sm.outputs.semantic_model import SemanticModel

T = TypeVar("T", covariant=True)
T1 = TypeVar("T1")


@dataclass
class Example(Generic[T]):
    id: str
    sms: list[SemanticModel]
    table: T

    def replace_table(self, table: T1) -> Example[T1]:
        return Example(id=self.id, sms=self.sms, table=table)


@dataclass
class FullTable:
    table: ColumnBasedTable
    context: Context
    links: Matrix[list[Link]]

    def nrows(self) -> int:
        return self.table.nrows()

    def select_rows(self, indices: list[int]) -> FullTable:
        """Select a subset of rows based on a boolean mask"""
        return FullTable(
            table=self.table.select_rows(indices),
            context=self.context,
            links=Matrix([self.links.data[i] for i in indices]),
        )

    def remove_empty_links(self) -> Self:
        return self.__class__(
            table=self.table,
            context=self.context,
            links=self.links.map(
                lambda links: [link for link in links if link.end > link.start]
            ),
        )

    def to_dict(self):
        return {
            "version": 2,
            "table": self.table.to_dict(),
            "context": self.context.to_dict(),
            "links": [
                [[link.to_dict() for link in cell] for cell in row]
                for row in self.links.data
            ],
        }

    @classmethod
    def from_dict(cls, obj: dict):
        version = obj["version"]
        if not (version == "1.2" or version == "1.1" or version == 2):
            raise ValueError(f"Unknown version: {version}")

        table = ColumnBasedTable.from_dict(obj["table"])
        if "links" in obj:
            links = Matrix(
                [
                    [[Link.from_dict(link) for link in cell] for cell in row]
                    for row in obj["links"]
                ]
            )
        else:
            links = Matrix.default(table.shape(), list)

        return cls(
            table=table,
            context=Context.from_dict(obj["context"]),
            links=links,
        )


@dataclass
class Dataset:
    location: Path

    @contextmanager
    def _open(self) -> Generator[Union[ZipPath, Path], None, None]:
        if self.is_zip_file():
            # it is a zip file
            with ZipFile(self.location, mode="r") as zf:
                root = ZipPath(zf)
                if self.description_dir(root).exists():
                    yield root

                subdirs = list(root.iterdir())
                if len(subdirs) == 1 and self.description_dir(subdirs[0]).exists():
                    yield subdirs[0]
                else:
                    raise ValueError("Invalid dataset format")
        else:
            yield self.location

    def description_dir(self, root: Union[Path, ZipPath]) -> Union[Path, ZipPath]:
        return root / "descriptions"

    def table_dir(self, root: Union[Path, ZipPath]) -> Union[Path, ZipPath]:
        return root / "tables"

    def is_zip_file(self):
        return self.location.name.endswith(".zip")

    def load(self) -> list[Example[FullTable]]:
        """Load dataset from a folder. Assuming the following structure:

        descriptions (containing semantic descriptions of tables)
        ├── <table_fs_id>
        │   ├── version.01[.json|.yml]
        │   ├── version.02[.json|.yml]
        │   └── ...
            or
        ├── <table_fs_id>[.json|.yml]
        ├── ...
        tables (containing list of tables, the type of table depends on )
        ├── <table_fs_id>[.json|.csv|.xlsx][.gz|.bz2|.lz4]
        ├── ...

        We also support compressing formats such as .zip.
        descriptions
        ├── part-<num>.zip (no nested version folders)
        │   ├── <table_fs_id>[.json|.yml]
        |   |   ...
        tables
        ├── part-<num>.zip
        │   ├── <table_fs_id>[.json|.csv|.xlsx]
        |   |   ...

        The description can be specified in either json or yaml format. For more information on
        how the semantic models are deserialized from the format, checkout the corresponding
        deserialization functions.

        Args:
            data_dir:

        Returns:
        """

        def deser_table(table_id: str, data: bytes, ext: str):
            if ext == ".json":
                return FullTable.from_dict(orjson.loads(data))
            if ext == ".csv":
                column_based_table = ColumnBasedTable.from_dataframe(
                    pd.read_csv(BytesIO(data)),
                    table_id=table_id,
                )
                return FullTable(
                    table=column_based_table,
                    context=Context(),
                    links=Matrix.default(column_based_table.shape(), list),
                )
            if ext == ".xlsx":
                column_based_table = ColumnBasedTable.from_dataframe(
                    pd.read_excel(BytesIO(data)),
                    table_id=table_id,
                )
                return FullTable(
                    table=column_based_table,
                    context=Context(),
                    links=Matrix.default(column_based_table.shape(), list),
                )

            raise ValueError(f"Unknown file type: {ext}")

        examples = []
        with self._open() as root:
            descdir = self.description_dir(root)
            tabledir = self.table_dir(root)

            for infile in sorted(tabledir.iterdir(), key=attrgetter("name")):
                suffixes = Path(infile.name).suffixes
                if infile.name.startswith(".") or len(suffixes) == 0:
                    continue

                if suffixes[-1] != ".zip":
                    example_id = infile.name[: -sum(len(x) for x in suffixes)]
                    table = deser_table(example_id, infile.read_bytes(), suffixes[0])

                    if descdir.exists():
                        if (descdir / example_id).exists():
                            desc_file = max(
                                (descdir / example_id).iterdir(),
                                key=lambda file: int(file.name.split(".")[1]),
                            )
                            assert desc_file is not None
                        else:
                            desc_file = descdir / f"{example_id}.json"
                            if not desc_file.exists():
                                desc_file = descdir / f"{example_id}.yml"
                                assert desc_file.exists()

                        if desc_file.name.endswith(".json"):
                            raw_sms = orjson.loads(desc_file.read_bytes())
                            sms = [SemanticModel.from_dict(sm) for sm in raw_sms]
                        else:
                            yaml = YAML()
                            raw = yaml.load(BytesIO(desc_file.read_bytes()))
                            ns = Namespace.from_prefix2ns(raw["prefixes"])
                            sms = [
                                SemanticModel.from_yaml_dict(sm, ns)
                                for sm in raw["models"]
                            ]
                    else:
                        sms = []

                    examples.append(
                        Example(id=table.table.table_id, sms=sms, table=table)
                    )
                else:
                    assert (
                        infile.name.endswith(".zip")
                        and isinstance(infile, Path)
                        and isinstance(descdir, Path)
                        and not self.is_zip_file()
                    ), "Must not be a zip file"
                    part: dict[str, FullTable] = {}
                    with ZipFile(infile, mode="r") as zf:
                        for file in zf.infolist():
                            if not file.filename.endswith(".json"):
                                continue

                            table_id = Path(file.filename).stem
                            with zf.open(file, mode="r") as f:
                                table = deser_table(
                                    table_id, f.read(), Path(file.filename).suffixes[0]
                                )
                            part[table_id] = table

                    if descdir.exists():
                        lst = []
                        with ZipFile(descdir / infile.name, mode="r") as zf:
                            for file in zf.infolist():
                                table_id = Path(file.filename).stem
                                if table_id not in part:
                                    continue

                                assert file.filename.endswith(".json")
                                with zf.open(file, mode="r") as f:
                                    sms = [
                                        SemanticModel.from_dict(sm)
                                        for sm in orjson.loads(f.read())
                                    ]
                                    lst.append(
                                        Example(
                                            id=table_id, sms=sms, table=part[table_id]
                                        )
                                    )
                    else:
                        lst = [
                            Example(id=table.table.table_id, sms=[], table=table)
                            for table in part.values()
                        ]
                    assert len(lst) == len(part)
                    examples.extend(lst)

            return examples

    def save(
        self,
        examples: list[Example[FullTable]],
        individual_table_compressed: Optional[Literal["gz", "bz2", "lz4"]] = None,
        batch_compressed: bool = False,
        batch_size: int = 100,
        table_fmt_indent: Literal[0, 2] = 0,
        clean_previous_data: bool = True,
    ):
        descdir = self.description_dir(self.location)
        tabledir = self.table_dir(self.location)
        assert (
            not self.is_zip_file()
            and isinstance(descdir, Path)
            and isinstance(tabledir, Path)
        )

        if descdir.exists() and clean_previous_data:
            shutil.rmtree(descdir)
        descdir.mkdir(parents=True, exist_ok=True)

        if tabledir.exists() and clean_previous_data:
            shutil.rmtree(tabledir)
        tabledir.mkdir(parents=True, exist_ok=True)

        if batch_compressed:
            for i, bexamples in enumerate(batch(batch_size, examples)):
                bexamples: list[Example[FullTable]]
                filename = f"part-{i:04d}.zip"
                with ZipFile(descdir / filename, "w") as dzf, ZipFile(
                    tabledir / filename, "w"
                ) as tzf:
                    for e in bexamples:
                        ename = get_friendly_fs_id(e.table.table.table_id) + ".json"
                        dzf.writestr(
                            ename,
                            data=orjson.dumps([sm.to_dict() for sm in e.sms]),
                        )
                        tzf.writestr(
                            ename,
                            data=orjson.dumps(e.table.to_dict()),
                        )
        else:
            for e in examples:
                filename = get_friendly_fs_id(e.table.table.table_id)
                (descdir / filename).mkdir(exist_ok=True)
                compressed_filename = (
                    filename + f".json.{individual_table_compressed}"
                    if individual_table_compressed is not None
                    else filename + ".json"
                )
                json.ser(
                    [sm.to_dict() for sm in e.sms],
                    descdir / filename / "version.01.json",
                    indent=2,
                )
                json.ser(
                    e.table, tabledir / compressed_filename, indent=table_fmt_indent
                )


def get_friendly_fs_id(id: str) -> str:
    if id.startswith("http://") or id.startswith("https://"):
        if id.find("dbpedia.org") != -1:
            return (
                slugify(
                    urlparse(id).path.replace("/resource/", "").replace("/", "_")
                ).replace("-", "_")
                + "_"
                + md5(id.encode()).hexdigest()
            )

        if id.find("wikipedia.org") != -1:
            return (
                slugify(
                    urlparse(id).path.replace("/wiki/", "").replace("/", "_")
                ).replace("-", "_")
                + "_"
                + md5(id.encode()).hexdigest()
            )

        raise NotImplementedError()
    return slugify(id.replace("/", "_"), lowercase=False).replace("-", "_")


class SampleableTable(Protocol):
    def nrows(self) -> int:
        ...

    def select_rows(self, indices: list[int]) -> Self:
        ...


ST = TypeVar("ST", bound=SampleableTable, covariant=True)


def sample_table_data(
    examples: Sequence[Example[ST]], n_rows: int, seed: Optional[int] = None
) -> Sequence[Example[ST]]:
    """Sample data from each table in examples.

    Args:
        examples: list of examples
        n_rows: number of rows to sample per table
        seed: random seed
    """
    rng = random.Random(seed)
    new_exs = []
    for ex in examples:
        tbl_nrows = ex.table.nrows()
        if tbl_nrows <= n_rows:
            new_exs.append(ex)
        else:
            indices = rng.sample(range(tbl_nrows), n_rows)
            new_exs.append(
                Example(id=ex.id, sms=ex.sms, table=ex.table.select_rows(indices))
            )

    return new_exs



from __future__ import annotations

import functools
import os
from pathlib import Path
from typing import Any, Callable, MutableMapping, Optional, Sequence, TypeVar, Union

import orjson
import serde.pickle

F = TypeVar("F", bound=Callable)


class CacheMethod:
    @staticmethod
    def single_object_arg(args, _kwargs):
        return id(args[0])

    @staticmethod
    def single_literal_arg(args, _kwargs):
        return args[0]

    @staticmethod
    def two_object_args(args, _kwargs):
        return (id(args[0]), id(args[1]))

    @staticmethod
    def three_object_args(args, _kwargs):
        return (id(args[0]), id(args[1]), id(args[2]))

    @staticmethod
    def auto_object_args(args, _kwargs):
        return tuple(x if isinstance(x, (str, int, bool)) else id(x) for x in args)

    @staticmethod
    def as_is_posargs(args, _kwargs):
        return args

    @staticmethod
    def as_is(args, kwargs):
        return (args, tuple(kwargs.items()))

    @staticmethod
    def as_json(args, kwargs):
        return orjson.dumps((args, kwargs))

    @staticmethod
    def selected_args(selection: Sequence[Union[int, str]]):
        selected_args = sorted([x for x in selection if isinstance(x, int)])
        selected_kwargs = sorted([x for x in selection if isinstance(x, str)])

        def fn(args, kwargs):
            sargs = []
            skwargs = {}
            for i in selected_args:
                if i < len(args):
                    sargs.append(args[i])
            for k in selected_kwargs:
                if k in kwargs:
                    skwargs[k] = kwargs[k]
            return orjson.dumps((sargs, skwargs))

        return fn

    @staticmethod
    def cache(
        key: Callable[[tuple, dict], Union[tuple, str, bytes, int]],
        cache_attr: str = "_cache",
    ) -> Callable[[F], F]:
        """Cache instance's method during its life-time.
        Note: Order of the arguments is important. Different order of the arguments will result in different cache key.
        """

        def wrapper_fn(func):
            fn_name = func.__name__

            @functools.wraps(func)
            def fn(self, *args, **kwargs):
                if not hasattr(self, cache_attr):
                    setattr(self, cache_attr, {})
                cache = getattr(self, cache_attr)
                k = (fn_name, key(args, kwargs))
                if k not in cache:
                    cache[k] = func(self, *args, **kwargs)
                return cache[k]

            return fn

        return wrapper_fn  # type: ignore


def cache_to_file(
    filepath: Union[Path, str],
    serialize: Callable[[Any, Path | str], None] = serde.pickle.ser,
    deserialize: Callable[[Path | str], Any] = serde.pickle.deser,
):
    def wrapper_fn(func):
        @functools.wraps(func)
        def fn():
            if os.path.exists(filepath):
                return deserialize(filepath)

            res = func()
            serialize(res, filepath)
            return res

        return fn

    return wrapper_fn


def cache_fn(
    cache: MutableMapping[bytes, Any],
    key: Optional[Callable[[str, tuple, dict], bytes]] = None,
) -> Callable[[F], F]:
    def wrapper_fn(func):
        fn_name = func.__name__

        @functools.wraps(func)
        def fn(self, *args, **kwargs):
            k = orjson.dumps((fn_name, args, kwargs))
            if k not in cache:
                cache[k] = func(self, *args, **kwargs)
            return cache[k]

        return fn

    return wrapper_fn  # type: ignore



import math
from typing import Dict, Callable, List, Mapping, Protocol, Iterable, Tuple
from dataclasses import dataclass
from sm.evaluation.sm_metrics import ScoringFn
from tqdm import tqdm


ItemParents = Dict[str, Iterable[str]]

MAX_ANCESTOR_DISTANCE = 5
MAX_DESCENDANT_DISTANCE = 3
INF_DISTANCE = 100


class ItemDistanceProtocol(Protocol):
    def get_distance(self, pred_item: str, target_item: str) -> int:
        """Get the distance between the predicted and target items. The distance is positive
        if the predicted item is an ancestor of the target item, and negative if the predicted
        item is a descendant of the target item.
        If the predicted item is not related to the target item, the distance can be choosed to be any number larger than MAX_ANCESTOR_DISTANCE or less than MAX_DESCENDANT_DISTANCE."""
        ...


@dataclass
class DictItemDistance:
    __slots__ = ["item2parents"]

    # mapping from the item to its parents and their distances
    # 1 is direct parent, 2 is grandparent, etc.
    item2parents: Mapping[str, Mapping[str, int]]

    def get_distance(self, pred_item: str, target_item: str) -> int:
        if pred_item == target_item:
            return 0
        if pred_item in self.item2parents[target_item]:
            return self.item2parents[target_item][pred_item]
        if target_item in self.item2parents[pred_item]:
            return -self.item2parents[pred_item][target_item]
        return INF_DISTANCE


class HierarchyScoringFn(ScoringFn):
    def __init__(
        self,
        item_distance: ItemDistanceProtocol,
    ):
        self.item_distance = item_distance

    def get_match_score(self, pred_item: str, target_item: str):
        if pred_item == target_item:
            return 1.0
        distance = self.item_distance.get_distance(pred_item, target_item)
        assert distance != 0.0
        if distance > 0:
            # pred_predicate is the ancestor of the target
            if distance > MAX_ANCESTOR_DISTANCE:
                return 0.0
            return math.pow(0.8, distance)
        if distance < 0:
            # pred_predicate is the descendant of the target
            distance = -distance
            if distance > MAX_DESCENDANT_DISTANCE:
                return 0.0
            return math.pow(0.7, distance)
        return 0.0

    @staticmethod
    def construct(
        items: List[str],
        get_item_parents: Callable[[str], Iterable[str]],
        get_item_uri: Callable[[str], str],
        verbose: bool = False,
    ):
        """Create a HierarchyScoringFn from a list of items

        Args:
            items: List of items to build the scoring function
            get_item_parents: Function that returns the parents of an item
            get_item_uri: Function that returns the URI of an item
        """
        item2parents = {item: {} for item in items}

        for item in tqdm(items, disable=not verbose):
            # a mapping of visited item to the distance that we encounter it
            visited = {}
            stack: List[Tuple[str, int]] = [
                (parent, 1) for parent in get_item_parents(item)
            ]
            while len(stack) > 0:
                node, distance = stack.pop()
                if node in visited and distance >= visited[node]:
                    # we have visited this node before and since last time we visit
                    # the previous route is shorter, so we don't need to visit it again
                    continue

                visited[node] = distance
                item2parents[item][node] = min(
                    distance, item2parents[item].get(node, float("inf"))
                )

                for parent in get_item_parents(node):
                    stack.append((parent, distance + 1))

        return HierarchyScoringFn(
            DictItemDistance(
                {
                    get_item_uri(k): {get_item_uri(k2): v2 for k2, v2 in v.items()}
                    for k, v in item2parents.items()
                }
            )
        )



from . import sm_metrics
from .cpa_cta_metrics import CTAEvalOutput, _cpa_transformation, _get_cta, cpa, cta
from .hierarchy_scoring_fn import HierarchyScoringFn
from .transformation import SemModelTransformation
from .utils import (
    PrecisionRecallF1,
    PrecisionRecallF1Protocol,
    PrecisionRecallF1Support,
)

__all__ = [
    "SemModelTransformation",
    "cpa",
    "cta",
    "HierarchyScoringFn",
    "sm_metrics",
    "PrecisionRecallF1",
    "PrecisionRecallF1Support",
    "PrecisionRecallF1Protocol",
    "CTAEvalOutput",
    "_cpa_transformation",
    "_get_cta",
]



from typing import Set
from sm.outputs.semantic_model import SemanticModel, ClassNode


class SemModelTransformation:
    @classmethod
    def replace_class_nodes_by_subject_columns(
        cls, sm: SemanticModel, id_props: Set[str]
    ):
        """Replace a class node by its subject column (if have). One class node must have maximum one subject column"""
        rm_nodes = []
        for cnode in sm.iter_nodes():
            if isinstance(cnode, ClassNode):
                inedges = sm.in_edges(cnode.id)
                outedges = sm.out_edges(cnode.id)
                id_edges = [
                    outedge for outedge in outedges if outedge.abs_uri in id_props
                ]
                if len(id_edges) == 0:
                    continue
                if len(id_edges) > 1:
                    raise Exception(
                        f"Assuming one class node only has one subject column. Node: {cnode.id} have {len(outedges)} subject columns: {outedges}"
                    )

                id_edge = id_edges[0]

                # update edges
                for inedge in inedges:
                    sm.remove_edge(inedge.id)
                    inedge.target = id_edge.target
                    sm.add_edge(inedge)
                for outedge in outedges:
                    sm.remove_edge(outedge.id)
                    outedge.source = id_edge.target
                    sm.add_edge(outedge)
                sm.remove_edge(id_edge.id)
                rm_nodes.append(cnode.id)
        for uid in rm_nodes:
            sm.remove_node(uid)

    @classmethod
    def remove_isolated_nodes(cls, sm: SemanticModel):
        rm_nodes = []
        for n in sm.iter_nodes():
            if sm.degree(n.id) == 0:
                rm_nodes.append(n.id)
        for uid in rm_nodes:
            sm.remove_node(uid)



from __future__ import annotations

from copy import deepcopy
from dataclasses import dataclass
from typing import Callable, Generic, Iterator, List, Tuple, TypeVar, overload

T = TypeVar("T")
T1 = TypeVar("T1")


@dataclass
class Matrix(Generic[T]):
    """Helper class to work with 2D array with Python's object."""

    data: List[List[T]]

    @overload
    @staticmethod
    def default(shp: Tuple[int, int], default: Callable[[], T]) -> Matrix[T]:
        pass

    @overload
    @staticmethod
    def default(shp: Tuple[int, int], default: T) -> Matrix[T]:
        pass

    @staticmethod
    def default(shp: Tuple[int, int], default: Callable[[], T] | T) -> Matrix[T]:
        if callable(default):
            return Matrix([[default() for _ in range(shp[1])] for _ in range(shp[0])])
        return Matrix([[default for _ in range(shp[1])] for _ in range(shp[0])])

    def shape(self) -> Tuple[int, int]:
        nrows = len(self.data)
        if nrows == 0:
            return 0, 0

        ncols = {len(row) for row in self.data}
        if len(ncols) > 1:
            raise ValueError("Matrix is not rectangular")
        return nrows, ncols.pop()

    def shallow_copy(self):
        return Matrix([row.copy() for row in self.data])

    def deep_copy(self):
        return Matrix([[deepcopy(item) for item in row] for row in self.data])

    @overload
    def __getitem__(self, item: int | Tuple[int, slice] | Tuple[slice, int]) -> List[T]:
        ...

    @overload
    def __getitem__(self, item: Tuple[int, int]) -> T:
        ...

    @overload
    def __getitem__(self, item: slice | Tuple[slice, slice]) -> List[List[T]]:
        ...

    def __getitem__(
        self,
        item: int
        | slice
        | Tuple[int, int]
        | Tuple[slice, slice]
        | Tuple[int, slice]
        | Tuple[slice, int],
    ) -> List[T] | List[List[T]] | T:
        if isinstance(item, (int, slice)):
            return self.data[item]
        if isinstance(item[0], slice):
            return [row[item[1]] for row in self.data[item[0]]]  # type: ignore
        return self.data[item[0]][item[1]]

    def get(self, key: Tuple[int, int], default: T) -> T:
        try:
            return self[key]
        except IndexError:
            return default

    def __setitem__(self, key: Tuple[int, int], value: T):
        self.data[key[0]][key[1]] = value

    def flat_iter(self) -> Iterator[T]:
        return (item for row in self.data for item in row)

    def enumerate_flat_iter(self) -> Iterator[Tuple[int, int, T]]:
        return (
            (ri, ci, item)
            for ri, row in enumerate(self.data)
            for ci, item in enumerate(row)
        )

    def map(self, fn: Callable[[T], T1]) -> Matrix[T1]:
        return Matrix([[fn(item) for item in row] for row in self.data])

    def map_index(self, fn: Callable[[int, int], T1]) -> Matrix[T1]:
        nrows, ncols = self.shape()
        return Matrix([[fn(ri, ci) for ci in range(ncols)] for ri in range(nrows)])



from sm.misc.fn_cache import *
from sm.misc.funcs import *
from sm.misc.bijection import Bijection, IntBijection
from sm.misc.matrix import Matrix


class UnreachableError(Exception):
    pass


__all__ = [
    "UnreachableError",
    "Bijection",
    "IntBijection",
    "Matrix",
]
